// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright (C) 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================


#ifndef FENNEC_LANG_LIMITS_H
#define FENNEC_LANG_LIMITS_H

#include <cmath>
#include <fennec/lang/types.h>
#include <fennec/lang/type_traits.h>


namespace fennec
{

///
/// \brief enum for determining rounding styles
enum float_round_style
{
	round_indeterminate       = -1
,	round_toward_zero         =  0
,	round_to_nearest          =  1
,	round_toward_infinity     =  2
,	round_toward_neg_infinity =  3
};

///
/// \brief Helper for getting traits of a numeric type
/// \tparam TypeT Numeric types, may be overloaded for custom types
template<typename TypeT> struct numeric_limits
{
	static constexpr bool is_specialized    = false;
	static constexpr bool is_signed         = false;
	static constexpr bool is_integer        = false;
	static constexpr bool is_exact          = false;
	static constexpr bool has_infinity      = false;
	static constexpr bool has_quiet_nan     = false;
	static constexpr bool has_signaling_nan = false;
	static constexpr bool has_denorm        = false;
	static constexpr bool has_denorm_loss   = false;
	static constexpr bool is_iec559         = false;
	static constexpr bool is_bounded        = false;
	static constexpr bool is_modulo         = false;
	static constexpr bool tinyness_before   = false;
	static constexpr bool traps             = false;

	static constexpr int  digits            = 0;
	static constexpr int  digits10          = 0;
	static constexpr int  max_digits10      = 0;
	static constexpr int  radix             = 0;
	static constexpr int  min_exponent      = 0;
	static constexpr int  min_exponent10    = 0;
	static constexpr int  max_exponent      = 0;
	static constexpr int  max_exponent10    = 0;

	static constexpr float_round_style rounding_style = round_indeterminate;

	static constexpr TypeT min()           { return TypeT(); }
	static constexpr TypeT max()           { return TypeT(); }
	static constexpr TypeT lowest()        { return TypeT(); }
	static constexpr TypeT epsilon()       { return TypeT(); }
	static constexpr TypeT round_error()   { return TypeT(); }
	static constexpr TypeT infinity()      { return TypeT(); }
	static constexpr TypeT quiet_NaN()     { return TypeT(); }
	static constexpr TypeT signaling_NaN() { return TypeT(); }
	static constexpr TypeT denorm_min()    { return TypeT(); }
};

// Overload for the builtin floating point type
template<> struct numeric_limits<float>
{
	static constexpr bool is_specialized    = true;
	static constexpr bool is_signed         = true;
	static constexpr bool is_integer        = false;
	static constexpr bool is_exact          = false;
	static constexpr bool has_infinity      = FLT_HAS_INFINITY;
	static constexpr bool has_quiet_nan     = FLT_HAS_QUIET_NAN;
	static constexpr bool has_signaling_nan = FLT_HAS_SIGNALING_NAN;
	static constexpr bool has_denorm        = FLT_HAS_DENORM;
	static constexpr bool has_denorm_loss   = FLT_HAS_DENORM_LOSS;
	static constexpr bool is_iec559         = FLT_IS_IEC559;
	static constexpr bool is_bounded        = true;
	static constexpr bool is_modulo         = false;
	static constexpr bool tinyness_before   = FLT_TINYNESS_BEFORE;
	static constexpr bool traps             = FLT_TRAPS;

	static constexpr int  digits            = FLT_MANT_DIG;
	static constexpr int  digits10          = FLT_DIG;
	static constexpr int  max_digits10      = FLT_DECIMAL_DIG;
	static constexpr int  radix             = FLT_RADIX;
	static constexpr int  min_exponent      = FLT_MIN_EXP;
	static constexpr int  min_exponent10    = FLT_MIN_10_EXP;
	static constexpr int  max_exponent      = FLT_MAX_EXP;
	static constexpr int  max_exponent10    = FLT_MAX_10_EXP;

	static constexpr double min()           { return  FLT_MIN;           }
	static constexpr double max()           { return  FLT_MAX;           }
	static constexpr double lowest()        { return -FLT_MAX;           }
	static constexpr double epsilon()       { return  FLT_EPSILON;       }
	static constexpr double round_error()   { return  FLT_ROUND_ERR;     }
	static constexpr double infinity()      { return  FLT_INF;           }
	static constexpr double quiet_NaN()     { return  FLT_QUIET_NAN;     }
	static constexpr double signaling_NaN() { return  FLT_SIGNALING_NAN; }
	static constexpr double denorm_min()    { return  FLT_DENORM_MIN;    }
};

// Overload for the bultin double precision floating point type
template<> struct numeric_limits<double>
{
	static constexpr bool is_specialized    = true;
	static constexpr bool is_signed         = true;
	static constexpr bool is_integer        = false;
	static constexpr bool is_exact          = false;
	static constexpr bool has_infinity      = DBL_HAS_INFINITY;
	static constexpr bool has_quiet_nan     = DBL_HAS_QUIET_NAN;
	static constexpr bool has_signaling_nan = DBL_HAS_SIGNALING_NAN;
	static constexpr bool has_denorm        = DBL_HAS_DENORM;
	static constexpr bool has_denorm_loss   = DBL_HAS_DENORM_LOSS;
	static constexpr bool is_iec559         = DBL_IS_IEC559;
	static constexpr bool is_bounded        = true;
	static constexpr bool is_modulo         = false;
	static constexpr bool tinyness_before   = DBL_TINYNESS_BEFORE;
	static constexpr bool traps             = DBL_TRAPS;

	static constexpr int  digits            = DBL_MANT_DIG;
	static constexpr int  digits10          = DBL_DIG;
	static constexpr int  max_digits10      = DBL_DECIMAL_DIG;
	static constexpr int  radix             = DBL_RADIX;
	static constexpr int  min_exponent      = DBL_MIN_EXP;
	static constexpr int  min_exponent10    = DBL_MIN_10_EXP;
	static constexpr int  max_exponent      = DBL_MAX_EXP;
	static constexpr int  max_exponent10    = DBL_MAX_10_EXP;

	static constexpr double min()           { return  DBL_MIN;           }
	static constexpr double max()           { return  DBL_MAX;           }
	static constexpr double lowest()        { return -DBL_MAX;           }
	static constexpr double epsilon()       { return  DBL_EPSILON;       }
	static constexpr double round_error()   { return  DBL_ROUND_ERR;     }
	static constexpr double infinity()      { return  DBL_INF;           }
	static constexpr double quiet_NaN()     { return  DBL_QUIET_NAN;     }
	static constexpr double signaling_NaN() { return  DBL_SIGNALING_NAN; }
	static constexpr double denorm_min()    { return  DBL_DENORM_MIN;    }
};

}

#endif // FENNEC_LANG_LIMITS_H
