// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright (C) 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================

///
/// \file matrix.h
/// \brief the \ref page_fennec_math_matrix
///
///
/// \details
/// \author Medusa Slockbower
///
/// \copyright Copyright © 2025 Medusa Slockbower ([GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html))
///
///

///
///
///
/// \page page_fennec_math_matrix Matrices
///
///
///
///

#ifndef FENNEC_MATH_MATRIX_H
#define FENNEC_MATH_MATRIX_H

#include <fennec/math/detail/__fwd.h>

#include <fennec/containers/array.h>
#include <fennec/math/geometric.h>
#include <fennec/math/vector.h>
#include <fennec/math/vector_traits.h>

namespace fennec
{


///
/// \brief returns a **copy** of the column \f$i\f$ of matrix \f$m\f$
/// \tparam scalar the base scalar type
/// \tparam rows number of rows in the matrix
/// \tparam cols indices of the columns in the matrix
/// \param m the matrix
/// \param i the index of the row
/// \returns a **copy** of the column at index \f$i\f$
template<typename scalar, size_t rows, size_t...cols>
constexpr vec<scalar, rows> column(const matrix<scalar, rows, cols...>& m, size_t i) noexcept
	{ return m[i]; }


///
/// \brief returns a **copy** of the row \f$i\f$ of matrix \f$m\f$
/// \tparam scalar the base scalar type
/// \tparam rows number of rows in the matrix
/// \tparam cols indices of the columns in the matrix
/// \param m the matrix
/// \param i the index of the row
/// \returns a **copy** of the row at index \f$i\f$
template<typename scalar, size_t rows, size_t...cols>
constexpr vec<scalar, sizeof...(cols)> row(const matrix<scalar, rows, cols...>& m, size_t i) noexcept
	{ return vec<scalar, sizeof...(cols)>(m[cols][i]...); }



template<size_t rows, size_t cols> using tmat = mat<float_t, rows, cols>;

using mat2 = tmat<2, 2>;
using mat3 = tmat<3, 3>;
using mat4 = tmat<4, 4>;

using mat2x2 = tmat<2, 2>;
using mat2x3 = tmat<2, 3>;
using mat2x4 = tmat<2, 4>;
using mat3x2 = tmat<3, 2>;
using mat3x3 = tmat<3, 3>;
using mat3x4 = tmat<3, 4>;
using mat4x2 = tmat<4, 2>;
using mat4x3 = tmat<4, 3>;
using mat4x4 = tmat<4, 4>;

template<size_t rows, size_t cols> using tdmat = mat<float_t, rows, cols>;

using dmat2 = tdmat<2, 2>;
using dmat3 = tdmat<3, 3>;
using dmat4 = tdmat<4, 4>;

using dmat2x2 = tdmat<2, 2>;
using dmat2x3 = tdmat<2, 3>;
using dmat2x4 = tdmat<2, 4>;
using dmat3x2 = tdmat<3, 2>;
using dmat3x3 = tdmat<3, 3>;
using dmat3x4 = tdmat<3, 4>;
using dmat4x2 = tdmat<4, 2>;
using dmat4x3 = tdmat<4, 3>;
using dmat4x4 = tdmat<4, 4>;


///
///
/// \brief
/// \tparam ScalarT
/// \tparam RowsV
/// \tparam ColIndicesV
template<typename ScalarT, size_t RowsV, size_t...ColIndicesV>
struct matrix
{
// Assertions ==========================================================================================================

	static_assert(is_arithmetic_v<ScalarT> or is_bool_v<ScalarT>);


// Typedefs & Constants ================================================================================================

	///
	/// \brief number of rows in the matrix
	static constexpr size_t rows = RowsV;

	///
	/// \brief number of columns in the matrix
	static constexpr size_t columns = sizeof...(ColIndicesV);

	///
	/// \brief number of total components
	static constexpr size_t num_components = rows * columns;

	///
	/// \brief vector class correspondant to the rows
	using row_t = vec<ScalarT, columns>;

	///
	/// \brief vector class correspondant to the columns
	using column_t = vec<ScalarT, rows>;

	///
	/// \brief base scalar type
	using scalar_t = ScalarT;

	///
	/// \brief alias for this matrix type
	using matrix_t = matrix;

	///
	/// \brief alias for the transpose matrix type
	using transpose_t = mat<ScalarT, columns, rows>;


	///
	/// \brief matrix array data
	array<column_t, columns> data;



// Constructors ========================================================================================================

	/// \name Constructors
	/// @{


	///
	/// \brief default constructor, initializes elements with 0
	///
	/// \details
	constexpr matrix()
		: data{ scalar_t(0) }{}


	///
	/// \brief copy constructor
	///
	/// \details
	/// \param mat matrix to copy
	constexpr matrix(const matrix_t& mat)
		: data{ mat.data } {}


	///
	/// \brief move constructor
	///
	/// \details
	/// \param mat matrix to move
	constexpr matrix(matrix_t&& mat)
		: data{ mat.data } {}


	///
	/// \brief scalar constructor, initializes a diagonal matrix with a scale of \p s
	///
	/// \details
	/// This function creates a diagonal matrix such that ```vec3(2.0f)``` would result in a matrix
	/// <table>
	/// <caption id="fennec_table_matrix_diagonal"></caption>
	/// <tr><th>   <th> 0 <th> 1 <th> 2
	/// <tr><th> 0 <td>2.0<td>0.0<td>0.0
	/// <tr><th> 1 <td>0.0<td>2.0<td>0.0
	/// <tr><th> 2 <td>0.0<td>0.0<td>2.0
	/// </table><br><br>
	///
	/// \param s scalar value
	///
	constexpr matrix(scalar_t s)
		{ (((ColIndicesV < rows) ? data[ColIndicesV][ColIndicesV] = s : 0), ...); }


	///
	/// \brief Piece-wise constructor
	///
	/// \details
	/// \tparam ArgsT
	/// \param args
	template<typename...ArgsT> requires(total_component_count_v<ArgsT...> == num_components)
	constexpr matrix(ArgsT&&...args)
		{ matrix::__construct(std::forward<ArgsT>(args)...); }

	/// @}

// Assignment Operators ================================================================================================

	/// \name Assignment Operators
	/// @{

	///
	/// \brief copy assignment operator
	///
	/// \details
	/// \returns a reference to **this** after copying the contents of \p rhs
	/// \param rhs the matrix to copy
	constexpr matrix_t& operator=(const matrix_t& rhs)
		{ data = rhs.data; return *this; }

	///
	/// \brief move assignment operator
	///
	/// \details
	/// \returns a reference to **this** after moving the contents of \p rhs
	/// \param rhs the matrix to move
	constexpr matrix_t& operator=(matrix_t&& rhs) noexcept
		{ data = rhs.data; return *this; }

	/// @}


// Access Operators ====================================================================================================

	/// \name Access Operator
	/// @{

	///
	/// \copydetails matrix::operator[](size_t) const
	constexpr column_t& operator[](size_t i)
		{ assert(i < rows); return data[i]; }

	///
	/// \brief returns the column at index \f$i\f$
	///
	/// \details
	/// \param i the index
	/// \returns the column at index \f$i\f$
	constexpr const column_t& operator[](size_t i) const
		{ assert(i < rows); return data[i]; }

	///
	/// \fn matrix::operator()(size_t, size_t)
	/// \copydetails matrix::operator()(size_t, size_t) const
	constexpr scalar_t& operator()(size_t i, size_t j)
		{ assert(i < rows && j < columns); return data[i][j]; }

	///
	/// \brief returns the cell in row \p j column \p i
	/// \param i the column
	/// \param j the row
	/// \returns the cell at the specified index.
	constexpr scalar_t operator()(size_t i, size_t j) const
		{ assert(i < rows && j < columns); return data[i][j]; }

	/// @}

// Scalar Operators ====================================================================================================

	/// \name Scalar Operators
	/// @{



	/// @}

// Vector Operators ====================================================================================================

	/// \name \ref matrix_t "matrix" \ref fennec::vector "vector" Operators
	/// @{

	///
	/// \brief performs a linear algebraic multiply
	/// \param lhs the matrix
	/// \param rhs the vector
	/// \returns a vector containing the dot products of \f$rhs\f$ with each row of \f$lhs\f$
	constexpr friend column_t operator*(const matrix_t& lhs, const row_t& rhs)
		{ return [lhs, rhs]<size_t...i>(index_sequence<i...>) -> column_t
		{ return column_t(fennec::dot(fennec::row(lhs, i), rhs) ...); }(make_index_sequence<rows>{}); }

	///
	/// \brief performs a linear algebraic multiply
	/// \param lhs the vector
	/// \param rhs the matrix
	/// \returns a vector containing the dot products of \f$lhs\f$ with each column of \f$rhs\f$
	constexpr friend row_t operator*(const column_t& lhs, const matrix_t& rhs)
		{ return row_t(fennec::dot(fennec::column(rhs, ColIndicesV), lhs) ...); }

	/// @}



// Matrix Operators ====================================================================================================

	/// \name \ref matrix_t "matrix" \ref matrix_t "matrix" Operators
	/// @{

	///
	/// \brief performs a linear algebraic matrix multiplication
	/// \param lhs the rows to multiply with
	/// \param rhs the columns to multiply with
	constexpr friend matrix_t operator*(const matrix_t& lhs, const transpose_t& rhs)
		{ return [lhs, rhs]<size_t...i>(index_sequence<i...>) -> matrix_t
		{ return matrix_t(rhs * fennec::row(lhs, i)...); }(make_index_sequence<rows>{}); }

	constexpr friend bool operator==(const matrix_t& lhs, const matrix_t& rhs)
		{ return lhs.data == rhs.data; }

	constexpr friend bool operator!=(const matrix_t& lhs, const matrix_t& rhs)
		{ return lhs.data != rhs.data; }

	/// @}



// Helpers =============================================================================================================

private:

	template<size_t i0 = 0>
	constexpr void __construct() { }

	template<size_t i0 = 0, typename HeadT, typename...ArgsT>
	constexpr void __construct(HeadT&& head, ArgsT&&...args)
		{ matrix::__insert(head); matrix::__construct<i0 + component_count_v<HeadT>>(std::forward<ArgsT>(args)...); }

	template<size_t i0 = 0>
	constexpr void __insert(scalar_t s)
		{ data[i0 / columns][i0 % columns] = s; }


	template<size_t i0 = 0, typename OScalarT> requires(is_arithmetic_v<OScalarT>)
	constexpr void __insert(OScalarT s)
		{ data[i0 / columns][i0 % columns] = scalar_t(s); }


	template<size_t i0 = 0, size_t...i>
	constexpr void __insert(const vector<scalar_t, i...>& v)
		{ (matrix::__insert<i0 + i>(v[i]), ...); }

};

}


#endif // FENNEC_MATH_MATRIX_H
